
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "cell.h"
#include "sif.h"

static void	Error(msg)
char		*msg;
{
	fputs(msg,stderr); fputc('\n',stderr); exit(1);
}

#define LINE_LEN	512
#define DESC_LEN	256

#define MA(num,ptr)	malloc((num)*sizeof(*(ptr)))

int	main(argc,argv)
int	argc;
char	**argv;
{
	FILE	*file;
	char	line[LINE_LEN],desc[DESC_LEN],
		*amae="array data memory allocation error.";
	int	N,M,n,m,bps;
	double	**array,min,max,base,step,aLim;
	Cell	cLim,**cell;

	if (argc!=4 && argc!=6)
	    Error("usage : a2tiff array {base step} BPS TIFF");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL)
		Error("array data file not found.");

	if (fgets(line,LINE_LEN,file)==NULL ||
	    sscanf(line,"%d %d",&N,&M)!=2 ||
	    N<=0 || M<=0) Error("bad number of array data.");

	fprintf(stderr,"%d\t%d\n",N,M);

	if ((array=(double **)MA(M,array))==NULL) Error(amae);

	for (m=0; m<M; m++) {
	    if ((array[m]=(double *)MA(N,*array))==NULL) Error(amae);

	    fread(array[m],sizeof(**array),N,file);
	    for (n=0; n<N; n++)
		if (m==0 && n==0)
		    min=max=array[0][0];
		else {
		    if (array[m][n]<min) min=array[m][n];
		    if (array[m][n]>max) max=array[m][n];
		}
	}
	if (file!=stdin) fclose(file);

	fprintf(stderr,"%lf\t%lf\n",min,max);

	if ((bps=atoi(argv[argc-2]))<=0 || bps>8*sizeof(Cell))
	    Error("bad image BPS.");

	cLim=(Cell)((1<<bps)-1);

	if (argc==4) {
	    base=min;
	    step=(max-min)/(double)cLim;

	    sprintf(desc,"%lf\t%lf\n",min,max);
	}
	else {
	    base=atof(argv[2]);
	    if ((step=atof(argv[3]))<=0.0) Error("bad value step.");

	    sprintf(desc,"%lf\t%lf\n%lf\t%lf",min,max,base,step);
	}
	aLim=base+step*(double)cLim;

	cell=(Cell **)array;
	for (m=0; m<M; m++) for (n=0; n<N; n++)
	    if (array[m][n]<=base)
		cell[m][n]=0;
	    else if (array[m][n]>=aLim)
		cell[m][n]=cLim;
	    else
		cell[m][n]=(Cell)floor((array[m][n]-base)/step+0.5);

	StoreImageFile(argv[argc-1],N,M,bps,cell,desc); return 0;
}
