
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

extern void	Error(char *msg);

static void	Dgemm(char OA,char OB,
		      int NAC,int NBC,int NAB,
		      double FAB,double *A,int LDA,double *B,int LDB,
		      double FC,double *C,int LDC)
{
	(void)dgemm_(&OA,&OB,&NAC,&NBC,&NAB,&FAB,A,&LDA,B,&LDB,&FC,C,&LDC);
}

#define LEN	2048

int	main(int argc,char **argv)
{
	double	*Q,*R,*F,*G,*H,*q,*r,*f,*g,*h,y1,y2,x1,x2,Dx=1.0,Dy=1.0;
	FILE	*in,*out;
	char	str[LEN];
	int	Nx,Ny,y,x,z=1;
	size_t	Nxy,n;

#ifdef	SPLINE
	if (argc!=3 && argc!=5)
	    Error("usage : dp2xp_spline DP.bin {Dx Dy} XP.bin");

	if (argc==5)
#else
	double	TOP;

	if (argc!=4 && argc!=6)
	    Error("usage : dp2xp DP.bin {Dx Dy} TOP XP.bin");

	TOP=atof(argv[argc-2]);
/*
	if (TOP<0.0 || TOP>1.0) Error("bad trade-off parameter.");
*/
	if (argc==6)
#endif
	if ((Dx=atof(argv[2]))<=0.0 ||
	    (Dy=atof(argv[3]))<=0.0) Error("bad detector interval.");

	if (argv[1][0]=='-')
	    in=stdin;
	else
	    if ((in=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,in)==NULL ||
	    sscanf(str,"%d %d %d",&Nx,&Ny,&z)<2 ||
	    Nx<=0 || Ny<=0 || z<=0) Error("bad projection data size.");

	Nxy=(size_t)Nx*Ny;

	if ((Q=(double *)malloc(sizeof(double)*Ny*Ny))==NULL ||
	    (R=(double *)malloc(sizeof(double)*Nx*Nx))==NULL ||
	    (F=(double *)malloc(sizeof(double)*Nxy))==NULL ||
	    (G=(double *)malloc(sizeof(double)*Nxy))==NULL ||
	    (H=(double *)malloc(sizeof(double)*Nxy))==NULL)
	    Error("memory allocation error.");

	if (argv[argc-1][0]=='-')
	    out=stdout;
	else
	    if ((out=fopen(argv[argc-1],"wb"))==NULL) Error("file not open.");

	(void)fprintf(out,"%d\t%d\t%d\n",Nx,Ny,z);

	for (q=Q, y=0; y<Ny; y++)
	for (x=0; x<Ny; x++, q++)
	    *q=sin(M_PI*((double)x+0.5)*(double)y/(double)Ny);

	for (r=R, y=0; y<Nx; y++)
	for (x=0; x<Nx; x++, r++)
	    *r=sin(M_PI*((double)x+0.5)*(double)y/(double)Nx);

	for (f=F, y=0; y<Ny; y++) { y1=(double)y/Dy/(double)Ny; y2=y1*y1;
	for (x=0; x<Nx; x++, f++) { x1=(double)x/Dx/(double)Nx; x2=x1*x1;
	    *f=(x==0 || y==0)?0.0:
#ifdef	SPLINE
		4.0/(double)Nxy/M_PI/M_PI*(x2+2.0*y2)/y2/y2;
#else
	       -4.0/(double)Nxy/M_PI/M_PI/(x2+TOP*y2);
#endif
	}}
	while (--z>=0) {
	    if (fread(G,sizeof(double),Nxy,in)!=Nxy)
		Error("projection data read error.");

	    for (g=G, h=H, y=0; y<Ny; y++)
	    for (x=0; x<Nx; x++, g++, h++) *h=(g[x!=Nx-1]-g[-(x!=0)])/2.0/Dx;

	    Dgemm('t','n',Nx,Ny,Nx,1.0,R,Nx,H,Nx,0.0,G,Nx);	/* G=(R^T) H */
	    Dgemm('n','n',Nx,Ny,Ny,1.0,G,Nx,Q,Ny,0.0,H,Nx);	/* H=G Q */

	    for (n=0; n<Nxy; n++) H[n]*=F[n];

	    Dgemm('n','n',Nx,Ny,Nx,1.0,R,Nx,H,Nx,0.0,G,Nx);	/* G=R H */
	    Dgemm('n','t',Nx,Ny,Ny,1.0,G,Nx,Q,Ny,0.0,H,Nx);	/* H=G (Q^T) */

	    (void)fwrite(H,sizeof(double),Nxy,out);
	}
	if (out!=stdout) (void)fclose(out);
	if (in !=stdin ) (void)fclose(in );

	free(H); free(G); free(F); free(R); free(Q);

	return 0;
}
