
#include <stdio.h>
#include <stdlib.h>
#include "sif_r.h"

extern void	Error(char *msg);

typedef struct RL {
			unsigned	i1,i2;
			struct RL	*next;
	} RL;

#define LIM	(unsigned)~0

static RL	*SetUp(char *arg)
{
	int	c;
	size_t	i1,i2;
	RL	*next,*rl=NULL;
	char	*brl="bad range list.";

#define GET(c)	(c=(*(arg++)))
#define S2I(i)	\
	for (i=0; c>='0' && c<='9'; GET(c)) if ((i=i*10+c-'0')>LIM) Error(brl)

	do {
	    while (GET(c)==',') ;

	    if (c!='\0') {
		if (c=='-') {
		    i1=0;
		    if (GET(c)==',' || c=='\0')
			i2=LIM;
		    else {
			S2I(i2);
			if (c!=',' && c!='\0') Error(brl);
		    }
		}
		else {
		    S2I(i1);
		    if (c==',' || c=='\0')
			i2=i1;
		    else if (c=='-')
			if (GET(c)==',' || c=='\0')
			    i2=LIM;
			else {
			    S2I(i2);
			    if ((c!=',' && c!='\0') || i2<i1) Error(brl);
			}
		    else
			Error(brl);
		}
		next=rl;
		if ((rl=(RL *)malloc(sizeof(RL)))==NULL)
		    Error("no allocatable memory for range list.");

		rl->i1=(unsigned)i1; rl->i2=(unsigned)i2; rl->next=next;
	    }
	} while (c==',');

	return rl;
}

static int	Include(RL *rl,unsigned i)
{
	int	e=2;

	for (; rl!=NULL; rl=rl->next)
	    if (i<=rl->i2) if (i>=rl->i1) return 1; else e=0;

	return e;
}

static unsigned	Image=0;

#define LEN	2048

static void	Abort(char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s at image %u.",msg,Image); Error(str);
}

static void	Skip(FILE *file,size_t size,char *msg)
{
	char	str[LEN];

	if (file!=stdin) {
	    if (fseeko(file,(off_t)size,SEEK_CUR)) goto abort;
	}
	else
	    do if (fgetc(file)==EOF) goto abort; while (--size!=0);

	return;
abort:;
	(void)sprintf(str,"%s skip error",msg); Abort(str);
}

#define SHORT	unsigned short

static unsigned	NumOfImages=0;
static SHORT	RemChannels=1;
static size_t	SNpz,Npz;
static int	Np,Nz;

static int	Header(FILE *file)
{
	char		header[64];
	unsigned	*NOI=(unsigned *)(header+14);
	SHORT		*CL=(SHORT *)(header+2),
			*IW=(SHORT *)(header+4),
			*IH=(SHORT *)(header+6),
			*FT=(SHORT *)(header+12),
			*RC=(SHORT *)(header+18);

	if (feof(file) || (NumOfImages==0 && RemChannels==0)) return 0;

	if (fread(&header,64,1,file)!=1 ||
	    header[0]!='I' || header[1]!='M') Abort("header read error");

	if (*FT!=2) Abort("unsupported file type");

	if (Image==0) {
	    if (*IW==0 || *IH==0) Abort("bad size");

	    SNpz=sizeof(SHORT)*(Npz=(Np=(*IW))*(Nz=(*IH)));
	}
	else
	    if (*IW!=Np || *IH!=Nz) Abort("size not match");

	if (*CL!=0) Skip(file,*CL,"comment");

	if (NumOfImages==0) {
	    NumOfImages=(*NOI); RemChannels=(*RC);
	}
	--NumOfImages;

	return 1;
}

static void	Read(FILE *file,SHORT *data,char *msg)
{
	char	str[LEN];

	if (fread(data,SNpz,1,file)!=1) {
	    (void)sprintf(str,"%s read error",msg); Abort(str);
	}
}

int	main(int argc,char **argv)
{
	FILE	*HIS;
	int	Ns,Nd,Ns1,p,s,z,v;
	RL	*rl;
	double	*K,*D,*d,sum,mom,i_d,ri1,ri2,rr1,rr2;
	SHORT	***I,*i;
	FOM	**RI,**RR,*ri,*rr;
	size_t	pz;
	char	path[LEN],*md="missing data";

	if (argc!=7 && argc!=10)
	    Error("usage : "
    "his2raw HIS_file scans viewsRangeList darks {A B C} RI_format RR_format"
		 );

	if (argv[1][0]=='-')
	    HIS=stdin;
	else
	    if ((HIS=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if ((Ns=atoi(argv[2]))<2) Error("bad number of scans.");

	rl=SetUp(argv[3]);
	
	if ((Nd=atoi(argv[4]))<1) Error("bad number of darks.");

	Ns1=Ns-1;

	NumOfImages=0;
	RemChannels=1;
	for (Image=0; Image<Nd; Image++) {
	    if (!Header(HIS)) Abort("missing dark data");

	    if (Image==0) {
		if ((K=(double *)malloc(sizeof(double)*Np))==NULL ||
		    (I=(SHORT ***)malloc(sizeof(SHORT **)*Ns1))==NULL ||
		    (*I=(SHORT **)malloc(sizeof(SHORT *)*Ns1*Nz))==NULL ||
		    (**I=(SHORT *)malloc(SNpz*Ns1))==NULL ||
		    (D=(double *)malloc(sizeof(double)*Npz))==NULL ||
		    (RI=(double **)malloc(sizeof(FOM *)*Nz))==NULL ||
		    (*RI=(double *)malloc(sizeof(FOM)*Nz*Ns1))==NULL ||
		    (RR=(double **)malloc(sizeof(FOM *)*Nz))==NULL ||
		    (*RR=(double *)malloc(sizeof(FOM)*Nz*Ns1))==NULL)
		    Error("memory allocation error.");

		if (argc!=10)
		    for (p=0; p<Np; p++) K[p]=1.0;
		else
{
	double	k,A=atof(argv[5]),B=atof(argv[6]),C=atof(argv[7]);

	for (p=0; p<Np; p++)
	    if ((K[p]=(A*(double)p+B)*(double)p+C)>0.0) {
		if (p==0 || K[p]>k) k=K[p];
	    }
	    else
		Error("bad coefficients for vignetting correction.");

	for (p=0; p<Np; p++) K[p]=k/K[p];
}
		for (s=0; s<Ns1; s++) {
		    if (s>0) I[s]=I[s-1]+Nz;

		    for (z=0; z<Nz; z++) if (s>0 || z>0) I[s][z]=I[s][z-1]+Np;
		}
		d=D; for (pz=0; pz<Npz; pz++) *(d++)=0.0;

		for (z=1; z<Nz; z++) {
		    RI[z]=RI[z-1]+Ns1; RR[z]=RR[z-1]+Ns1;
		}
	    }
	    Read(HIS,i=I[0][0],"dark data");
	    d=D; for (pz=0; pz<Npz; pz++) *(d++)+=((double)*(i++)/(double)Nd);
	}
	for (v=0; Header(HIS); v++) {
	    Skip(HIS,SNpz,"junk data"); ++Image;

	    switch(Include(rl,v)) { 
		case 0:
		    for (s=0; s<Ns1; s++) {
			if (!Header(HIS)) Abort(md);

			Skip(HIS,SNpz,"data"); ++Image;
		    }
		    break;
		case 1:
		    for (s=0; s<Ns1; s++) {
			if (!Header(HIS)) Abort(md);

			Read(HIS,I[s][0],"data"); ++Image;
		    }
		    for (ri=(*RI), rr=(*RR), d=D, z=0; z<Nz; z++, d+=Np)
		    for (s=0; s<Ns1; s++, ri++, rr++) {
			i=I[s][z]; sum=mom=0.0;
			for (p=0; p<Np; p++)
			    if ((i_d=K[p]*((double)i[p]-d[p]))>0.0) {
				sum+=i_d; mom+=(i_d*(double)p);
			    }

			if (sum>0.0) {
			    *ri=sum/(double)Np; *rr=mom/sum;
			}
			else
			    *ri=(*rr)=0.0;

			if (s==0 && z==0) {
			    ri1=ri2=(*ri);
			    rr1=rr2=(*rr);
			}
			else {
			    if (ri1>(*ri)) ri1=(*ri); else
			    if (ri2<(*ri)) ri2=(*ri);
			    if (rr1>(*rr)) rr1=(*rr); else
			    if (rr2<(*rr)) rr2=(*rr);
			}
		    }
		    (void)printf("%d\t%lf\t%lf\t%lf\t%lf\n",v,ri1,ri2,rr1,rr2);

		    if (argv[argc-2][0]!='-') {
			(void)sprintf(path,argv[argc-2],v);
			StoreImageFile_Raw(path,Ns1,Nz,ri1,ri2,RI);
		    }
		    if (argv[argc-1][0]!='-') {
			(void)sprintf(path,argv[argc-1],v);
			StoreImageFile_Raw(path,Ns1,Nz,rr1,rr2,RR);
		    }
		    break;
		default:
		    goto Break;
	    }
	}
Break:;
	free(*RR); free(RR);
	free(*RI); free(RI); free(D); free(**I); free(*I); free(I); free(K);

	if (HIS!=stdin) (void)fclose(HIS);

	return 0;
}
