
#include <stdio.h>
#include <stdlib.h>

extern void	Error(char *msg);

static unsigned	Image;

#define LEN	2048

static void	Abort(char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s at image %u.",msg,Image); Error(str);
}

static void	Skip(FILE *file,size_t size,char *msg)
{
	char	str[LEN];

	if (file!=stdin) {
	    if (fseeko(file,(off_t)size,SEEK_CUR)) goto abort;
	}
	else
	    do if (fgetc(file)==EOF) goto abort; while (--size!=0);

	return;
abort:;
	(void)sprintf(str,"%s skip error",msg); Abort(str);
}

static size_t	SNpz,Npz;
static int	Np,Nz;

#define SHORT	unsigned short

static void	Header(FILE *file)
{
	char	header[64];
	SHORT	*CL=(SHORT *)(header+2),
		*IW=(SHORT *)(header+4),
		*IH=(SHORT *)(header+6),
		*FT=(SHORT *)(header+12);

	if (fread(&header,64,1,file)!=1 ||
	    header[0]!='I' || header[1]!='M') Abort("header read error");

	if (*FT!=2) Abort("unsupported file type");

	if (Image==0) {
	    if (*IW==0 || *IH==0) Abort("bad size");

	    SNpz=sizeof(SHORT)*(Npz=(Np=(*IW))*(Nz=(*IH)));
	}
	else
	    if (*IW!=Np || *IH!=Nz) Abort("size not match");

	if (*CL!=0) Skip(file,*CL,"comment");
}

static void	Read(FILE *file,SHORT *data)
{
	if (fread(data,SNpz,1,file)!=1) Abort("data read error");
}

static void	Raw(FILE *HIS,int Ns1,int Nv,SHORT ***I,double *D,char *path)
{
	FILE	*file;
	char	str[LEN];
	int	s,z,p;
	double	*d,w,m,t;
	SHORT	*i;

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) {
		(void)sprintf(str,"%s : file not open.",path); Error(str);
	    }

	(void)fprintf(file,"%d\t%d\t%d\n",Ns1,Nz,Nv);

	while (--Nv>=0) {
	    Header(HIS); Skip(HIS,SNpz,"junk data"); ++Image;

	    for (s=0; s<Ns1; s++) {
		Header(HIS); Read(HIS,I[s][0]); ++Image;
	    }
	    for (d=D, z=0; z<Nz; z++, d+=Np)
		for (s=0; s<Ns1; s++) {
		    i=I[s][z]; w=m=0.0;
		    for (p=0; p<Np; p++) {
			t=(double)i[p]-d[p]; w+=t; m+=(t*(double)p);
		    }
		    t=(w<=0.0)?0.0:-m/w; (void)fwrite(&t,sizeof(double),1,file);
		}
	}
	if (file!=stdout) (void)fclose(file);
}

int	main(int argc,char **argv)
{
	FILE	*HIS;
	int	Nd,Ns,Nv,Ns1,s,z;
	SHORT	***I,*i;
	double	*D,*d;
	size_t	pz;

	if (argc!=6 && argc!=7)
	    Error("usage : his2rr HIS.bin darks scans views raw.bin\n"
		  "        his2rr HIS.bin darks scans views 0.bin 180.bin");

	if (argv[1][0]=='-')
	    HIS=stdin;
	else
	    if ((HIS=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if ((Nd=atoi(argv[2]))<1) Error("bad number of darks.");
	if ((Ns=atoi(argv[3]))<2) Error("bad number of scans.");
	if ((Nv=atoi(argv[4]))<2) Error("bad number of views.");

	Ns1=Ns-1;

	for (Image=0; Image<Nd; Image++) {
	    Header(HIS);
	    if (Image==0) {
		if ((I=(SHORT ***)malloc(sizeof(SHORT **)*Ns1))==NULL ||
		    (*I=(SHORT **)malloc(sizeof(SHORT *)*Ns1*Nz))==NULL ||
		    (**I=(SHORT *)malloc(SNpz*Ns1))==NULL ||
		    (D=(double *)malloc(sizeof(double)*Npz))==NULL)
		    Error("memory allocation error.");

		for (s=0; s<Ns1; s++) {
		    if (s>0) I[s]=I[s-1]+Nz;

		    for (z=0; z<Nz; z++) if (s>0 || z>0) I[s][z]=I[s][z-1]+Np;
		}
		d=D; for (pz=0; pz<Npz; pz++) *(d++)=0.0;
	    }
	    Read(HIS,i=I[0][0]);
	    d=D; for (pz=0; pz<Npz; pz++) *(d++)+=((double)*(i++)/(double)Nd);
	}
	if (argc==6)
	    Raw(HIS,Ns1,Nv-1,I,D,argv[5]);
	else {
	    Raw(HIS,Ns1,1,I,D,argv[5]);

	    while (--Nv>=2)
		for (s=0; s<=Ns1; s++) {
		    Header(HIS); Skip(HIS,SNpz,"unused data"); ++Image;
		}
		
	    Raw(HIS,Ns1,1,I,D,argv[6]);
	}
	free(D); free(**I); free(*I); free(I);

	if (HIS!=stdin) (void)fclose(HIS);

	return 0;
}
