
#if	defined(LEFT) && defined(RIGHT)
#error both LEFT and RIGHT are defined.
#endif

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

static unsigned	Image;

#define LEN	2048

static void	Abort(char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s at image %u.",msg,Image); Error(str);
}

static void	Skip(FILE *file,size_t size,char *msg)
{
	char	str[LEN];

	if (file!=stdin) {
	    if (fseeko(file,(off_t)size,SEEK_CUR)) goto abort;
	}
	else
	    do if (fgetc(file)==EOF) goto abort; while (--size!=0);

	return;
abort:;
	(void)sprintf(str,"%s skip error",msg); Abort(str);
}

static size_t	Npz;
static int	Np,Nz;

#define SHORT	unsigned short

static void	Header(FILE *file)
{
	char	header[64];
	SHORT	*CL=(SHORT *)(header+2),
		*IW=(SHORT *)(header+4),
		*IH=(SHORT *)(header+6),
		*FT=(SHORT *)(header+12);

	if (fread(&header,64,1,file)!=1 ||
	    header[0]!='I' || header[1]!='M') Abort("header read error");

	if (*FT!=2) Abort("unsupported file type");

	if (Image==0) {
	    if (*IW==0 || *IH==0) Abort("bad size");

	    Npz=(Np=(*IW))*(Nz=(*IH));
	}
	else
	    if (*IW!=Np || *IH!=Nz) Abort("size not match");

	if (*CL!=0) Skip(file,*CL,"comment");
}

static void	Read(FILE *file,SHORT *data)
{
	if (fread(data,sizeof(SHORT),Npz,file)!=Npz) Abort("data read error");
}

static FILE	*Open(char *path,int Ns,int Nz,int Nv)
{
	FILE	*file;
	char	str[LEN];

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) {
		(void)sprintf(str,"%s : file not open.",path); Error(str);
	    }

	(void)fprintf(file,"%d\t%d\t%d\n",Ns,Nz,Nv); return file;
}

static void	Profile(SHORT *I,double *D,double *SOP,double *COP)
{
	int	p;
	double	i,s=0.0,m=0.0;

	for (p=0; p<Np; p++) {
	    i=(double)*(I++)-(*(D++)); s+=i; m+=(i*(double)p);
	}
	if (s<=0.0)
	    *SOP=(*COP)=0.0;
	else {
	    *SOP=s; *COP=m/s;
	}
}

static double	Log(double num,double den)
{
	return ((num>0.0)?log(num):0.0)-((den>0.0)?log(den):0.0);
}

static void	Write(FILE *file,double xp)
{
	(void)fwrite(&xp,sizeof(double),1,file);
}

static void	Split(FILE *HIS,
		      int N0,int Ns1,int Nv,SHORT ***I,double *D,
		      char *axp,char *rxp)
{
	FILE	*AXP=Open(axp,Ns1,Nz,Nv),
		*RXP=Open(rxp,Ns1,Nz,Nv);
	size_t	SNpz=sizeof(SHORT)*Npz;
	int	s,z;
	double	*d,sop,cop;
#ifndef	LEFT
	SHORT	***i=I+Ns1-1;
#endif
#if	defined(LEFT) || defined(RIGHT)
	double	S0,C0;
#else
	int	ds=Ns1-1-(N0-1);
	double	S1,C1,S2,C2,r2,r1,s0=(double)(N0-1)/2.0;
#endif
	while (--Nv>=0) {
	    Header(HIS); Skip(HIS,SNpz,"data"); ++Image;

	    for (s=0; s<Ns1; s++) {
		Header(HIS); Read(HIS,I[s][0]); ++Image;
	    }
	    for (d=D, z=0; z<Nz; z++, d+=Np) {
#if	defined(LEFT) || defined(RIGHT)
		S0=C0=0.0;
#else
		S1=C1=S2=C2=0.0;
#endif
		for (s=0; s<N0; s++) {
#ifndef	RIGHT
		    Profile(I[ s][z],d,&sop,&cop);
#ifdef	LEFT
		    S0+=(sop/(double)N0); C0+=(cop/(double)N0);
#else
		    S1+=(sop/(double)N0); C1+=(cop/(double)N0);
#endif
#endif
#ifndef	LEFT
		    Profile(i[-s][z],d,&sop,&cop);
#ifdef	RIGHT
		    S0+=(sop/(double)N0); C0+=(cop/(double)N0);
#else
		    S2+=(sop/(double)N0); C2+=(cop/(double)N0);
#endif
#endif
		}
		for (s=0; s<Ns1; s++) {
		    Profile(I[s][z],d,&sop,&cop);
#if	defined(LEFT) || defined(RIGHT)
		    Write(AXP,Log(S0,sop));
#ifdef	CR
		    Write(RXP,cop-C0);
#else
		    Write(RXP,C0-cop);
#endif
#else
		    r2=1.0-(r1=((double)s-s0)/(double)ds);
		    Write(AXP,Log(r2*S1+r1*S2,sop));
#ifdef	CR
		    Write(RXP,cop-C2*r1-C1*r2);
#else
		    Write(RXP,r2*C1+r1*C2-cop);
#endif
#endif
		}
	    }
	}
	if (RXP!=stdout) (void)fclose(RXP);
	if (AXP!=stdout) (void)fclose(AXP);
}

int	main(int argc,char **argv)
{
	FILE	*HIS;
	int	Nd,N0,Ns,Nv,Ns1,s,z;
	size_t	pz;
	SHORT	***I,*i;
	double	*D,*d;

	if (argc!=8 && argc!=10)
	    Error(
"usage : his2xp HIS.bin darks I0s scans views aXP.bin rXP.bin {aXP_180 rXP_180}"
	    );

	if (argv[1][0]=='-')
	    HIS=stdin;
	else
	    if ((HIS=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if ((Nd=atoi(argv[2]))<1) Error("bad number of darks.");
	if ((N0=atoi(argv[3]))<1) Error("bad number of I0s at both sides.");
	if ((Ns=atoi(argv[4]))<N0
#if	!defined(LEFT) && !defined(RIGHT)
	    *2
#endif
	    +2) Error("bad number of scans.");
	if ((Nv=atoi(argv[5]))<2) Error("bad number of views.");

	Ns1=Ns-1;

	for (Image=0; Image<Nd; Image++) {
	    Header(HIS);
	    if (Image==0) {
		if ((I=(SHORT ***)malloc(sizeof(SHORT **)*Ns1))==NULL ||
		    (*I=(SHORT **)malloc(sizeof(SHORT *)*Ns1*Nz))==NULL ||
		    (**I=(SHORT *)malloc(sizeof(SHORT)*Ns1*Npz))==NULL ||
		    (D=(double *)malloc(sizeof(double)*Npz))==NULL)
		    Error("memory allocation error.");

		for (s=0; s<Ns1; s++) {
		    if (s>0) I[s]=I[s-1]+Nz;

		    for (z=0; z<Nz; z++) if (s>0 || z>0) I[s][z]=I[s][z-1]+Np;
		}
		d=D; for (pz=0; pz<Npz; pz++) *(d++)=0.0;
	    }
	    Read(HIS,i=I[0][0]);
	    d=D; for (pz=0; pz<Npz; pz++) *(d++)+=((double)*(i++)/(double)Nd);
	}
	Split(HIS,N0,Ns1,Nv-1,I,D,argv[6],argv[7]);

	if (argc==10) Split(HIS,N0,Ns1,1,I,D,argv[8],argv[9]);

	free(D); free(**I); free(*I); free(I);

	if (HIS!=stdin) (void)fclose(HIS);

	return 0;
}
