
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#ifndef	Filter
#define Filter	Chesler
#endif

#ifndef M_PI
#define M_PI	3.14159265358979323846
#endif

extern void	Error(char *msg);
extern double	Filter(int j);

double	Ramachandran(int j)
{
	return (j==0)?1.0/4.0:(j&1)?-1.0/(M_PI*M_PI*(double)j*(double)j):0.0;
}

double	Shepp(int j)
{
	return 2.0/(M_PI*M_PI*(1.0-4.0*(double)j*(double)j));
}

double	Chesler(int j)
{
	if (j==0)
	    return 1.0/8.0-1.0/(2.0*M_PI*M_PI);
	else if (j==1 || j==(-1))
	    return 1.0/16.0-1.0/(2.0*M_PI*M_PI);
	else if (j&1)
	    return -1.0/(2.0*M_PI*M_PI*(double)j*(double)j);
{
	double  j2=(double)j*(double)j,j21=j2-1.0;

	return -(j2+1.0)/(2.0*M_PI*M_PI*j21*j21);
}}


#ifdef	_WIN32
#include <process.h>
#include <windows.h>

#define THREAD_T	HANDLE
#define FUNCTION_T	unsigned __stdcall
#define RETURN_VALUE	0

#define INIT_THREAD(T,F,A)	\
	(T=(HANDLE)_beginthreadex(NULL,0,F,(void *)A,0,NULL))==0
#define TERM_THREAD(T)	\
	WaitForSingleObject(T,INFINITE)==WAIT_FAILED || CloseHandle(T)==0
#else
#include <pthread.h>

#define THREAD_T	pthread_t
#define FUNCTION_T	void *
#define RETURN_VALUE	NULL

#define INIT_THREAD(T,F,A)	pthread_create(&(T),NULL,F,(void *)A)
#define TERM_THREAD(T)		pthread_join(T,NULL)
#endif

static int	Threads=1;

typedef struct {
		double	x,y;
	} Vector;

static int	Nr,Nz,N1,N2;
static double	**PQ,factor;
static Vector	*E,*G,**V;

static void	FFT(int sign,Vector *GV)
{
	int	j,k,l,m,p,q;
	Vector	gv,*e,*a,*b;
	double	x,y,z;

	for (j=N2-1, k=N2-2; k>0; k--) {
	    for (l=N1; (j^=l)&l; l>>=1) ;

	    if (j>k) {
		gv=GV[j]; GV[j]=GV[k]; GV[k]=gv;
	    }
	}
	for (j=2, k=1, l=N1, m=N2; m>1; m=l, l>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=m) {
		x=e->x; y=e->y*(double)sign;
		for (a=GV+p, q=0; q<l; q++, a+=j) {
		    gv=(*(b=a+k)); b->x=a->x-(z=gv.x*x-gv.y*y); a->x+=z;
				   b->y=a->y-(z=gv.x*y+gv.y*x); a->y+=z;
		}
	    }
}

static FUNCTION_T	Convolution(void *i)
{
	size_t	l;
	double	*pq,x,y;
	int	k,j;
	Vector	*v=V[(size_t)i];

	for (l=(size_t)i; l<Nz; l+=Threads) {
	    pq=PQ[l];

	    for (k=0; k<N1; k++) v[k].x=v[k].y=0.0;

	    for (j=0; j<Nr; j++, k++) {
		v[k].x=factor*pq[j]; v[k].y=0.0;
	    }
	    for (; k<N2; k++) v[k].x=v[k].y=0.0;

	    FFT(-1,v);

	    for (k=0; k<N2; k++) {
		x=v[k].x;
		y=v[k].y; v[k].x=G[k].x*x-G[k].y*y;
			  v[k].y=G[k].x*y+G[k].y*x;
	    }
	    FFT(1,v);

	    for (j=0; j<Nr; j++) pq[j]=v[j].x;
	}
	return RETURN_VALUE;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	char		*env,str[LEN];
	FILE		*in,*out;
	int		Nt,j,i,l,m;
	double		x,y,Dr=1.0;
	THREAD_T	*Thread;

	if (argc!=3 && argc!=4) Error("usage : xp2fp XP {Dr} FP");

	if ((env=getenv("THREADS"))!=NULL &&
	    (Threads=atoi(env))<=0) Error("bad number of THREADS.");

	if (argv[1][0]=='-')
	    in=stdin;
	else
	    if ((in=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,in)==NULL ||
	    sscanf(str,"%d %d %d",&Nr,&Nz,&Nt)!=3 ||
	    Nr<=0 || Nz<=0 || Nt<=0) Error("bad projection data size.");

	if (argc==4)
	    if ((Dr=atof(argv[2]))<=0.0) Error("bad detector interval.");

	for (N1=1; N1<Nr; N1<<=1) ; N2=N1<<1;

	if ((Thread=(THREAD_T *)malloc(sizeof(THREAD_T)*Threads))==NULL ||
	    (E=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (G=(Vector *)malloc(sizeof(Vector)*N2))==NULL ||
	    (V=(Vector **)malloc(sizeof(Vector *)*Threads))==NULL ||
	    (*V=(Vector *)malloc(sizeof(Vector)*Threads*N2))==NULL ||
	    (PQ=(double **)malloc(sizeof(double *)*Nz))==NULL ||
	    (*PQ=(double *)malloc(sizeof(double)*Nz*Nr))==NULL)
	    Error("memory allocation error.");

	if (argv[argc-1][0]=='-')
            out=stdout;
        else
            if ((out=fopen(argv[argc-1],"wb"))==NULL) Error("file not open.");

	(void)fprintf(out,"%d\t%d\t%d\n",Nr,Nz,Nt);

	factor=1.0/Dr/(double)N2;

	for (j=0; j<N2; j++) {
	    sincos(M_PI*(double)j/(double)N2,&(E[j].y),&(E[j].x));

	    G[j].x=Filter(j-N1); G[j].y=0.0;
	}
	FFT(-1,G);

	for (i=1; i<Threads; i++) V[i]=V[i-1]+N2;

	for (l=1; l<Nz; l++) PQ[l]=PQ[l-1]+Nr;

	for (m=0; m<Nt; m++) {
	    if (fread(*PQ,sizeof(double),(size_t)Nz*Nr,in)!=(size_t)Nr*Nz)
		Error("projection data read error.");

	    for (i=1; i<Threads; i++)
		if (INIT_THREAD(Thread[i],Convolution,(void *)(size_t)i))
		    Error("multi-threading initialization error.");

	    (void)Convolution((void *)0);

	    for (i=1; i<Threads; i++)
		if (TERM_THREAD(Thread[i]))
		    Error("multi-threading termination error.");

	    (void)fwrite(*PQ,sizeof(double),(size_t)Nz*Nr,out);
	}
	if (out!=stdout) (void)fclose(out);
	if (in !=stdin ) (void)fclose(in );

	free(*PQ); free(PQ); free(*V); free(V); free(G); free(E); free(Thread);

	return 0;
}
