
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

#define LEN	2048

static void	Store(char *path,int Nx,int Ny,double *P)
{
	FILE	*file;
	char	str[LEN];

	if (*path=='-')
	    file=stdout;
	else
	    if ((file=fopen(path,"wb"))==NULL) {
		(void)sprintf(str,"%s : file not open.",path); Error(str);
	    }

	(void)fprintf(file,"%d\t%d\n",Nx,Ny);
	(void)fwrite(P,sizeof(double),(size_t)Nx*Ny,file);
	if (file!=stdout) (void)fclose(file);
}

int	main(int argc,char **argv)
{
	FILE	*file;
	char	str[LEN];
	int	Nx,Ny,z,Nz=1;
	size_t	Nxy,xy;
	double	*P1,*P2,p;

	if (argc!=4) Error("usage : xp2msd XP.bin XP_mean XP_SD.bin");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d %d",&Nx,&Ny,&Nz)<2 ||
	    Nx<=0 || Ny<=0 || Nz<=0) Error("bad projection data size.");

	Nxy=(size_t)Nx*Ny;

	if ((P1=(double *)malloc(sizeof(double)*Nxy))==NULL ||
	    (P2=(double *)malloc(sizeof(double)*Nxy))==NULL)
	    Error("memory allocation error.");

	for (xy=0; xy<Nxy; xy++) P1[xy]=P2[xy]=0.0;

	for (z=0; z<Nz; z++)
	    for (xy=0; xy<Nxy; xy++) {
		if (fread(&p,sizeof(double),1,file)!=1)
		    Error("projection data read error.");

		P1[xy]+=p;
		P2[xy]+=(p*p);
	    }

	if (file!=stdin) (void)fclose(file);

	for (xy=0; xy<Nxy; xy++) {
	    p=(P1[xy]/=(double)Nz);
	    P2[xy]=sqrt(P2[xy]/(double)Nz-p*p);
	}
	Store(argv[2],Nx,Ny,P1);
	Store(argv[3],Nx,Ny,P2);

	free(P2); free(P1);

	return 0;
}
