
#include <stdio.h>
#include <stdlib.h>
#include <math.h>

extern void	Error(char *msg);

#ifndef	M_PI
#define	M_PI	3.14159265358979323846
#endif

static int	Range(int *Ng,int *Nh,int *D1,int *D2,int *Og,int *Oh)
{
	int	g,h,N,L;

	if ((g=(*Nh)+(*D2))>(*Ng)) g=(*Ng);
	if ((h=(*Ng)-(*D1))>(*Nh)) h=(*Nh);

	if (*D2<0) {
	    *Og=0; *Oh=(-(*D2)); *D1-=(*D2); *D2=0; N=1-(*D1);
	}
	else if (*D1>0) {
	    *Og=(*D1); *Oh=0; *D2-=(*D1); *D1=0; N=(*D2)+1;
	}
	else {
	    *Og=(*Oh)=0; N=(-(*D1)>(*D2))?1-(*D1):*D2+1;
	}
	N+=(((*Ng=g-(*Og))>(*Nh=h-(*Oh)))?*Ng:*Nh);

	for (L=1; L<N; L<<=1) ; return L;
}

typedef struct {
		double	r,i;
	} Complex;

static void	FFT(int S,int L,Complex *E,Complex *C)
{
	int	j,k,m,n,p,q,L2=L>>1;
	Complex	c,*e,*a,*b;
	double	r,i,d;

	for (j=L-1, k=L-2; k>0; k--) {
	    for (m=L2; (j^=m)&m; m>>=1) ;

	    if (j>k) {
		c=C[j]; C[j]=C[k]; C[k]=c;
	    }
	}
	for (j=2, k=1, m=L2, n=L; n>1; n=m, m>>=1, k=j, j<<=1)
	    for (e=E, p=0; p<k; p++, e+=n) {
		r=e->r;
		i=e->i*(double)S;
		for (a=C+p, q=0; q<m; q++, a+=j) {
		    c=(*(b=a+k)); b->r=a->r-(d=c.r*r-c.i*i); a->r+=d;
				  b->i=a->i-(d=c.r*i+c.i*r); a->i+=d;
		}
	    }
}

static Complex	Product(double r1,double i1,double r2,double i2)
{
	Complex	c;

	c.r=r1*r2-i1*i2;
	c.i=r1*i2+i1*r2; return c;
}

double	*RMSD(int Ngx,int Ngy,double **G,
	      int Nhx,int Nhy,double **H,int Dx1,int Dy1,int Dx2,int Dy2)
{
	int	Lx,Ogx,Ohx,Lx2,x,Lx1,Ngx1,Ngx2,Nhx1,Nhx2,gx,hx,
		Ly,Ogy,Ohy,Ly2,y,Ly1,Ngy1,Ngy2,Nhy1,Nhy2,gy,hy,Lxx,z;
	double	*P,d,**Gy,**Hy,*Gx,*Hx,gr,gi,hr,hi,t,r,i,*p,Lxy,Lxyy;
	Complex	*Ex,*Ey,**Q,*U,*V,*ex,*ey,*q,*u,*v,g,h,c;

	Lx=Range(&Ngx,&Nhx,&Dx1,&Dx2,&Ogx,&Ohx); Lx2=Lx*2;
	Ly=Range(&Ngy,&Nhy,&Dy1,&Dy2,&Ogy,&Ohy); Ly2=Ly*2;

	if ((P=(double *)malloc(sizeof(double)*(Dy2-Dy1+1)*
					       (Dx2-Dx1+1)))==NULL ||
	    (Ex=(Complex *)malloc(sizeof(Complex)*Lx2))==NULL ||
	    (Ey=(Complex *)malloc(sizeof(Complex)*Ly2))==NULL ||
	    (Q=(Complex **)malloc(sizeof(Complex *)*Ly))==NULL ||
	    (*Q=(Complex *)malloc(sizeof(Complex)*Ly*Lx))==NULL ||
	    (U=(Complex *)malloc(sizeof(Complex)*Ly))==NULL ||
	    (V=(Complex *)malloc(sizeof(Complex)*Ly))==NULL) return NULL;

	for (ex=Ex, x=0; x<Lx; x++, ex++) {
	    d=M_PI*(double)x/(double)Lx; ex[Lx].r=(-(ex->r=cos(d)));
					 ex[Lx].i=(-(ex->i=sin(d)));
	}
	for (ey=Ey, y=0; y<Ly; y++, ey++) {
	    d=M_PI*(double)y/(double)Ly; ey[Ly].r=(-(ey->r=cos(d)));
					 ey[Ly].i=(-(ey->i=sin(d)));

	    if (y>0) Q[y]=Q[y-1]+Lx;
	}
	Lx1=Lx-1; Ngx1=Ngx-1; Nhx1=Nhx-1; Ngx2=Ngx%2; Nhx2=Nhx%2;
	Ly1=Ly-1; Ngy1=Ngy-1; Nhy1=Nhy-1; Ngy2=Ngy%2; Nhy2=Nhy%2;
/*
    ----
*/
	for (Gy=G+Ogy, Hy=H+Ohy, y=0; y<Ly; y++) {
	    x=0; q=Q[y];
	    if (y<Ngy)
		for (Gx=(*(Gy++))+Ogx; x<Ngx; x++) {
		    d=(*(Gx++)); (q++)->r=d*d;
		}

	    for (; x<Lx; x++) (q++)->r=0.0;

	    x=0; q=Q[y];
	    if (y<Nhy)
		for (Hx=(*(Hy++))+Ohx; x<Nhx; x++) {
		    d=(*(Hx++)); (q++)->i=d*d;
		}

	    for (; x<Lx; x++) (q++)->i=0.0;

	    if (y<Ngy || y<Nhy) FFT(-1,Lx,Ex,Q[y]);
	}
/*
    x=0
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][0];

	FFT(-1,Ly,Ey,U);

	U->r=U->r*(double)Nhx*(double)Nhy+(double)Ngx*(double)Ngy*U->i;
	U->i=0.0;
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=U+Ly1;
	     u<v;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--) {
	    t=ey->r/ey->i; r=Ey[hy].r;
			   i=Ey[hy].i; d=(double)Nhx*(r+i*t);
				       g=Product(u->r+v->r,u->i-v->i,d*r, d*i);
			   r=Ey[gy].r;
			   i=Ey[gy].i; d=(double)Ngx*(r+i*t);
				       h=Product(d*r,-d*i,u->i+v->i,u->r-v->r);
	    v->r=u->r=(g.r+h.r)/2.0;
		 u->i=(g.i+h.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*(double)Nhx*(double)Nhy2+(double)Ngx*(double)Ngy2*u->i;
	u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][0]=(*(u++));
/*
    x=1~Lx/2-1
*/
	for (ex=Ex+1, gx=Ngx1, hx=Nhx1, x=1, Lxx=Lx1;
	     x<Lxx;
	     ex++, gx=(gx+Ngx1)%Lx2, hx=(hx+Nhx1)%Lx2, x++, Lxx--) {
	    for (u=U, v=V, y=0; y<Ly; y++) {
		*(u++)=Q[y][x]; *(v++)=Q[y][Lxx];
	    }
	    FFT(-1,Ly,Ey,U); FFT(-1,Ly,Ey,V);

	    t=ex->r/ex->i; r=Ex[hx].r; i=Ex[hx].i; d=r+i*t; hr=d*r; hi=d*i;
			   r=Ex[gx].r; i=Ex[gx].i; d=r+i*t; gr=d*r; gi=d*i;

	    g=Product(U->r+V->r,U->i-V->i,hr*(double)Nhy, hi*(double)Nhy);
	    h=Product(gr*(double)Ngy,-gi*(double)Ngy,U->i+V->i,U->r-V->r);
	    V->r=U->r=(g.r+h.r)/2.0;
		 U->i=(g.i+h.i)/2.0; V->i=(-U->i);

	    for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=V+Ly1, y=1;
		 y<Ly;
		 ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--, y++) {
		t=ey->r/ey->i;
		r=Ey[hy].r;
		i=Ey[hy].i; d=r+i*t; c=Product(hr,hi,d*r,d*i);
				     g=Product(u->r+v->r,u->i-v->i,c.r, c.i);
		r=Ey[gy].r;
		i=Ey[gy].i; d=r+i*t; c=Product(gr,gi,d*r,d*i);
				     h=Product(c.r,-c.i,u->i+v->i,u->r-v->r);
		v->r=u->r=(g.r+h.r)/2.0;
		     u->i=(g.i+h.i)/2.0; v->i=(-u->i);
	    }
	    FFT(1,Ly,Ey,U); FFT(1,Ly,Ey,V);

	    for (u=U, v=V, y=0; y<Ly; y++) {
		Q[y][x]=(*(u++)); Q[y][Lxx]=(*(v++));
	    }
	}
/*
    x=Lx/2
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][x];

	FFT(-1,Ly,Ey,U);

	U->r=U->r*(double)Nhx2*(double)Nhy+(double)Ngx2*(double)Ngy*U->i;
	U->i=0.0;
	for (ey=Ey+1, gy=Ngy1, hy=Nhy1, u=U+1, v=U+Ly1;
	     u<v;
	     ey++, gy=(gy+Ngy1)%Ly2, hy=(hy+Nhy1)%Ly2, u++, v--) {
	    t=ey->r/ey->i;
	    r=Ey[hy].r;
	    i=Ey[hy].i; d=(double)Nhx2*(r+i*t);
			g=Product(u->r+v->r,u->i-v->i,d*r, d*i);
	    r=Ey[gy].r;
	    i=Ey[gy].i; d=(double)Ngx2*(r+i*t);
			h=Product(d*r,-d*i,u->i+v->i,u->r-v->r);
	    v->r=u->r=(g.r+h.r)/2.0;
		 u->i=(g.i+h.i)/2.0; v->i=(-u->i);
	}
	u->r=u->r*(double)Nhx2*(double)Nhy2+(double)Ngx2*(double)Ngy2*u->i;
	u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][x]=(*(u++));
/*
    ----
*/
	for (p=P, y=Dy1; y<=Dy2; y++) {
	    FFT(1,Lx,Ex,q=Q[(y<0)?Ly+y:y]);

	    for (x=Dx1; x<=Dx2; x++) *(p++)=q[(x<0)?Lx+x:x].r;
	}
/*
    ----
*/
	for (Gy=G+Ogy, Hy=H+Ohy, y=0; y<Ly; y++) {
	    x=0; q=Q[y];
	    if (y<Ngy)
		for (Gx=(*(Gy++))+Ogx, x=0; x<Ngx; x++) (q++)->r=(*(Gx++));

	    for (; x<Lx; x++) (q++)->r=0.0;

	    x=0; q=Q[y];
	    if (y<Nhy)
		for (Hx=(*(Hy++))+Ohx, x=0; x<Nhx; x++) (q++)->i=(*(Hx++));

	    for (; x<Lx; x++) (q++)->i=0.0;

	    if (y<Ngy || y<Nhy) FFT(-1,Lx,Ex,Q[y]);
	}
/*
    x=0
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][0];

	FFT(-1,Ly,Ey,U);

	U->r*=U->i; U->i=0.0;
	for (u=U+1, v=U+Ly1; u<v; u++, v--) {
	    *u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
		       (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							 v->i=(-u->i);
	}
	u->r*=u->i; u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][0]=(*(u++));
/*
    x=1~Lx/2-1
*/
	for (x=1, Lxx=Lx1; x<Lxx; x++, Lxx--) {
	    for (u=U, v=V, y=0; y<Ly; y++) {
		*(u++)=Q[y][x]; *(v++)=Q[y][Lxx];
	    }
	    FFT(-1,Ly,Ey,U); FFT(-1,Ly,Ey,V);

	    *U=Product((U->r+V->r)/2.0,(U->i-V->i)/2.0,
		       (U->i+V->i)/2.0,(U->r-V->r)/2.0); V->r=  U->r ;
							 V->i=(-U->i);
	    for (u=U+1, v=V+Ly1, y=1; y<Ly; y++, u++, v--) {
		*u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
			   (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							     v->i=(-u->i);
	    }
	    FFT(1,Ly,Ey,U); FFT(1,Ly,Ey,V);

	    for (u=U, v=V, y=0; y<Ly; y++) {
		Q[y][x]=(*(u++)); Q[y][Lxx]=(*(v++));
	    }
	}
/*
    x=Lx/2
*/
	for (u=U, y=0; y<Ly; y++) *(u++)=Q[y][x];

	FFT(-1,Ly,Ey,U);

	U->r*=U->i; U->i=0.0;
	for (u=U+1, v=U+Ly1; u<v; u++, v--) {
	    *u=Product((u->r+v->r)/2.0,(u->i-v->i)/2.0,
		       (u->i+v->i)/2.0,(u->r-v->r)/2.0); v->r=  u->r ;
							 v->i=(-u->i);
	}
	u->r*=u->i; u->i=0.0;

	FFT(1,Ly,Ey,U);

	for (u=U, y=0; y<Ly; y++) Q[y][x]=(*(u++));
/*
    ----
*/
	Lxy=(double)Lx*(double)Ly;
	for (p=P, y=Dy1; y<=Dy2; y++) {
	    if (y<0) {
		q=Q[Ly+y]; Lxyy=Lxy*(double)(((z=Nhy+y)<Ngy)?z:Ngy);
	    }
	    else {
		q=Q[   y]; Lxyy=Lxy*(double)(((z=Ngy-y)<Nhy)?z:Nhy);
	    }
	    FFT(1,Lx,Ex,q);

	    for (x=Dx1; x<=Dx2; x++, p++) {
		if (x<0)
		    d=(*p-2.0*q[Lx+x].r)/(double)(((z=Nhx+x)<Ngx)?z:Ngx);
		else
		    d=(*p-2.0*q[   x].r)/(double)(((z=Ngx-x)<Nhx)?z:Nhx);

		*p=(d>0.0)?sqrt(d/Lxyy):0.0;
	    }
	}
/*
    ----
*/
	free(V); free(U); free(*Q); free(Q); free(Ey); free(Ex); return P;
}

#define LEN	2048

static void	Abort(char *path,char *msg)
{
	char	str[LEN];

	(void)sprintf(str,"%s : %s",path,msg); Error(str);
}

int	main(int argc,char **argv)
{
	double	Rx,Ry,**G,**H,d,*Z,*z;
	FILE	*f0,*f180,*file;
	char	str[LEN],
		*fnf="projection image file not found.",
		*dre="projection image data read error.";
	int	i,Nx,Ny,Nz,y1,y2,y,x,Nx1,Dx,Dy,x0,y0;
	size_t	Nxy;

	switch(argc) {
	    case  5:
	    case  6: Rx=atof(argv[3]); Ry=atof(argv[4]); break;
	    case  7:
	    case  8: Rx=atof(argv[5]); Ry=atof(argv[6]); break;
	    default:
		Error("usage : xp2rc"
#ifdef	DPC
				    "_dpc"
#endif
				    " 0.bin 180.bin {y1 y2} Rx Ry {RMSD.txt}");
	}
	if (Rx<=0.0 || Rx>=1.0 ||
	    Ry<=0.0 || Ry>=1.0) Error("bad distance ratio.");

	if (argv[1][0]=='-')
	    f0=stdin;
	else
	    if ((f0=fopen(argv[1],"rb"))==NULL) Abort(argv[1],fnf);

	if (argv[2][0]=='-')
	    f180=stdin;
	else
	    if ((f180=fopen(argv[2],"rb"))==NULL) Abort(argv[2],fnf);

	if (fgets(str,LEN,f0)==NULL ||
	    (i=sscanf(str,"%d %d %d",&Nx,&Ny,&Nz))<2 ||
	    Nx<=0 || Ny<=0) Abort(argv[1],"bad projection image size.");

	if (argc==5 || argc==6) {
	    y1=0; y2=Ny-1;
	}
	else
	    if ((y1=atoi(argv[3]))<0 ||
		(y2=atoi(argv[4]))>=Ny || y1>y2) Error("bad slice range.");

	Nxy=Nx*Ny;

	if ((G=(double **)malloc(sizeof(double *)*Ny))==NULL ||
	    (*G=(double *)malloc(sizeof(double)*Nxy))==NULL ||
	    (H=(double **)malloc(sizeof(double *)*Ny))==NULL ||
	    (*H=(double *)malloc(sizeof(double)*Nxy))==NULL)
	    Error("no allocatable memory for projection image.");

	for (y=0; y<Ny; y++) {
	    if (y>0) {
		G[y]=G[y-1]+Nx;
		H[y]=H[y-1]+Nx;
	    }
	    if (fread(G[y],sizeof(double),Nx,f0)!=Nx) Abort(argv[1],dre);
	}
	if (f0==stdin && f180==stdin && i==3)
	    while (--Nz>0)
		if (fread(*H,sizeof(double),Nxy,f0)!=Nxy) Abort(argv[1],dre);

	if (f0!=stdin) (void)fclose(f0);

	if (fgets(str,LEN,f180)==NULL ||
	    sscanf(str,"%d %d",&x,&y)!=2 ||
	    x!=Nx || y!=Ny) Abort(argv[2],"projection image size not match.");

	Nx1=Nx-1;

	for (y=0; y<Ny; y++)
	for (x=Nx1; x>=0; x--) {
	    if (fread(&d,sizeof(double),1,f180)!=1) Abort(argv[2],dre);
#ifdef	DPC
	    H[y][x]=(-d);
#else
	    H[y][x]=d;
#endif
	}
	if (f180!=stdin) (void)fclose(f180);

	Dx=(int)(Rx*(double)Nx);
	Dy=(int)(Ry*(double)(y2-=(y1-1)));

	if ((Z=RMSD(Nx,y2,G+y1,Nx,y2,H+y1,-Dx,-Dy,Dx,Dy))==NULL)
	    Error("no allocatable memory for RMSD calculation.");

	if (argc==6 || argc==8)
	    if (argv[argc-1][0]=='-')
		file=stdout;
	    else
		if ((file=fopen(argv[argc-1],"w"))==NULL)
		    Abort(argv[argc-1],"file not open.");

	for (z=Z, y=(-Dy); y<=Dy; y++)
	for (x=(-Dx); x<=Dx; x++, z++) {
	    if (z==Z || *z<d) {
		x0=x; y0=y; d=(*z);
	    }
	    if (argc==6 || argc==8) (void)fprintf(file,"%d\t%d\t%lf\n",x,y,*z);
	}
	if ((argc==6 || argc==8) && file!=stdout) (void)fclose(file);

	free(Z); free(*H); free(H); free(*G); free(G);

	(void)printf("%lg\t%lg\t%lg\n"
		     "%d\t%d\t%d\n",(double)(Nx1-Dx)/2.0,
				    (double)(Nx1+Dx)/2.0,
				    (double)(Nx1+x0)/2.0,-Dy,Dy,y0); return 0;
}
