#!/bin/csh
# usage : xp2rc.csh     0.bin 180.bin {y1 y2} Rx Ry {RMSD.txt}
# usage : xp2rc_dpc.csh 0.bin 180.bin {y1 y2} Rx Ry {RMSD.txt}
#
set xp2rc=`basename $0`
#
#set rmsd=rmsd_2d
set rmsd=rmsd_2d.CC
#
#set tmp=/tmp/${xp2rc).$$
set tmp=/dev/shm/${xp2rc}.$$
#
alias STDERR 'echo echo \!* 1\>\&2 | /bin/bash'
#
switch($xp2rc)
  case xp2rc.csh :
    @ sign =  1 ; breaksw
  case xp2rc_dpc.csh :
    @ sign = -1 ; breaksw
  default :
    STDERR $xp2rc : unknown command. ; exit 1
endsw
#
switch($#argv)
  case [4,5] :
    set rxy=($argv[3-4]) ; breaksw
  case [6,7] :
    set rxy=($argv[5-6]) ; breaksw
  default :
    STDERR "usage : $xp2rc 0.bin 180.bin {y1 y2} Rx Ry {RMSD.txt}" ; exit 1
endsw
#
if (`echo $rxy | awk '{ print ($1<=0 || $1>=1 || $2<=0 || $2>=1) }'`) then
  STDERR bad ratio of distance. ; exit 1
endif
#
cp /dev/null ${tmp}.0
foreach index (1 2)
  a2tiff $argv[$index] 16 ${tmp}.$index >>& ${tmp}.0
  if ($status) then
    STDERR $argv[$index] : `tail -1 ${tmp}.0` ; rm -f ${tmp}.? ; exit 1
  endif
end
#
set nxy=(`sed -n -e 1p -e 3p ${tmp}.0 | uniq`)
if ($#nxy != 2) then
  STDERR image size not match. ; rm -f ${tmp}.? ; exit 1
endif
#
if ($#argv =~ [6,7]) then
  if ($argv[3] < 0 || $argv[4] >= $nxy[2] || $argv[3] > $argv[4]) then
    STDERR bad range of slice. ; rm -f ${tmp}.? ; exit 1
  endif
#
  set area=(0 $argv[3] `expr $nxy[1] - 1` $argv[4])
  trim_gray ${tmp}.1 $area ${tmp}.1
  trim_gray ${tmp}.2 $area ${tmp}.2
  @ nxy[2] = ( $argv[4] - $argv[3] ) + 1
endif
#
rar_gray ${tmp}.2 -x +y ${tmp}.2
#
if ($#argv =~ [4,6]) then
  alias RMSD '$rmsd \!*'
else
  alias RMSD '$rmsd \!* | tee $argv[$#argv]'
endif
#
set awk='{ printf "%0.16f %0.16f\n",S*$1,S*($2-$1)/65535 }'
#
set dxy=(`echo $nxy $rxy | \
          awk '{ x=int($1*$3); y=int($2*$4); print -x,-y,x,y }'`)
#
( echo $nxy[1] $dxy ; \
  RMSD ${tmp}.1 `awk -v S=1     "NR==2 $awk" ${tmp}.0` \
       ${tmp}.2 `awk -v S=$sign "NR==4 $awk" ${tmp}.0` $dxy | sort -n -k3 ) | \
awk 'NR==1 { X=$1-1; x1=(X+$2)/2; y1=$3; x2=(X+$4)/2; y2=$5 } \
     NR==2 { printf "%g\t%g\t%g\n%s\t%s\t%s\n",x1,x2,(X+$1)/2,y1,y2,$2 }'
#
rm -f ${tmp}.?
