
#include <stdio.h>
#include <stdlib.h>
#include "cell.h"
#include "sif.h"

extern void	Error(char *msg);

typedef struct RL {
			unsigned	i1,i2;
			struct RL	*next;
	} RL;

#define LIM	(unsigned)~0

static RL	*SetUp(char *arg)
{
	int	c;
	size_t	i1,i2;
	RL	*next,*rl=NULL;
	char	*brl="bad range list.";

#define GET(c)	(c=(*(arg++)))
#define S2I(i)	\
	for (i=0; c>='0' && c<='9'; GET(c)) if ((i=i*10+c-'0')>LIM) Error(brl)

	do {
	    while (GET(c)==',') ;

	    if (c!='\0') {
		if (c=='-') {
		    i1=0;
		    if (GET(c)==',' || c=='\0')
			i2=LIM;
		    else {
			S2I(i2);
			if (c!=',' && c!='\0') Error(brl);
		    }
		}
		else {
		    S2I(i1);
		    if (c==',' || c=='\0')
			i2=i1;
		    else if (c=='-')
			if (GET(c)==',' || c=='\0')
			    i2=LIM;
			else {
			    S2I(i2);
			    if ((c!=',' && c!='\0') || i2<i1) Error(brl);
			}
		    else
			Error(brl);
		}
		next=rl;
		if ((rl=(RL *)malloc(sizeof(RL)))==NULL)
		    Error("no allocatable memory for range list.");

		rl->i1=(unsigned)i1; rl->i2=(unsigned)i2; rl->next=next;
	    }
	} while (c==',');

	return rl;
}

static int	Include(RL *rl,unsigned i)
{
	int	e=2;

	for (; rl!=NULL; rl=rl->next)
	    if (i<=rl->i2) if (i>=rl->i1) return 1; else e=0;

	return e;
}

#define LEN	2048

int	main(int argc,char **argv)
{
	FILE	*file;
	RL	*rl,*next;
	char	str[LEN],mm[LEN];
	int	Nr,Nz,z,BPS,t,r,Nt=1;
	double	*P,*p,base,step,roof,min,max;
	size_t	sdNzNr;
	Cell	**C,c;

	if (argc!=3 && argc!=5 && argc!=7)
	Error("usage : xp2tiff XP.bin rangeList\n"
	      "        xp2tiff XP.bin rangeList {base step} BPS TIFF_format");

	if (argv[1][0]=='-')
	    file=stdin;
	else
	    if ((file=fopen(argv[1],"rb"))==NULL) Error("file not found.");

	if ((rl=SetUp(argv[2]))==NULL) Error("empty range list.");

	if (fgets(str,LEN,file)==NULL ||
	    sscanf(str,"%d %d %d",&Nr,&Nz,&Nt)<2 ||
	    Nr<=0 || Nz<=0 || Nt<=0) Error("bad projection data size.");

	sdNzNr=sizeof(double)*Nz*Nr;

	if ((P=(double *)malloc(sdNzNr))==NULL)
	    Error("no allocatable memory for projection data.");

	if (argc!=3) {
	    if ((C=(Cell **)malloc(sizeof(Cell *)*Nz))==NULL)
		Error("no allocatable memory for cell pointer.");

	    for (p=P, z=0; z<Nz; z++, p+=Nr) C[z]=(Cell *)p;

	    if ((BPS=atoi(argv[argc-2]))<=0 || BPS>8*sizeof(Cell))
		Error("bad image BPS (bits per sample.");

	    c=(Cell)((1<<BPS)-1);

	    if (argc!=5) {
		base=atof(argv[3]);
		if ((step=atof(argv[4]))<=0.0) Error("bad value step.");

		(void)sprintf(mm,"%lf\t%lf",base,roof=base+step*(double)c);
	    }
	}

#define PDE(s_r)	\
	(void)sprintf(str,"projection data " s_r " error at %d.",t); Error(str)

	for (t=0; t<Nt; t++)
	    switch(Include(rl,t)) {
		case 0:
		    if (file!=stdin) {
			if (!fseeko(file,(off_t)(sdNzNr),SEEK_CUR)) break;
		    }
		    else
			if (fread(P,sdNzNr,1,file)==1) break;

		    PDE("skip");
		case 1:
		    if (fread(P,sdNzNr,1,file)!=1) {
			PDE("read");
		    }
		    min=max=(*P);
		    for (p=P, z=0; z<Nz; z++)
		    for (r=0; r<Nr; r++, p++)
			if (*p<min) min=(*p); else if (*p>max) max=(*p);

		    (void)printf("%d\t%lf\t%lf\n",t,min,max);

		    if (argc!=3) {
			if (argc==5) {
			    step=(min<max)?(max-min)/(double)c:1.0;

			    (void)sprintf(mm,"%lf\t%lf",base=min,roof=max);
			}
			for (p=P, z=0; z<Nz; z++)
			for (r=0; r<Nr; r++, p++)
			    C[z][r]=(*p<=base)?0:
				    (*p>=roof)?c:
				    (Cell)((*p-base)/step+0.5);

			(void)sprintf(str,argv[argc-1],t);
			StoreImageFile(str,Nr,Nz,BPS,C,mm);
		    }
		    break;
		default:
		    goto Break;
	    }

Break:;
	if (argc!=3) free(C);

	free(P);

	for (; rl!=NULL; rl=next) {
	    next=rl->next; free(rl);
	}
	if (file!=stdin) (void)fclose(file);

	return 0;
}
